# -*- coding: utf-8 -*- #
"""Module to handle the importation from MRIFileManager and its progress

Contains:
    Class:
        -ImportProgress : Inherit from QProgressDialog and handle the
        progress bar
        -ImportWorker : Inherit from QThread and manage the threads
    Methods:
        -read_log : Show the evolution of the progress bar and returns its
        feedback
        -tags_from_file : Returns a list of [tag, value] contained in a Json
        file
        -verify_scans : Check if the project's scans have been modified


"""

##########################################################################
# Populse_mia - Copyright (C) IRMaGe/CEA, 2018
# Distributed under the terms of the CeCILL license, as published by
# the CEA-CNRS-INRIA. Refer to the LICENSE file or to
# http://www.cecill.info/licences/Licence_CeCILL_V2.1-en.html
# for details.
##########################################################################

import glob
import os.path
import json
import hashlib  # To generate the md5 of each path
import datetime
from time import time, sleep
from datetime import datetime
import threading

# PyQt5 imports
from PyQt5.QtCore import Qt, QThread, pyqtSignal
from PyQt5.QtWidgets import QProgressDialog

# Populse_MIA imports
from populse_mia.data_manager.project import (
    COLLECTION_CURRENT, COLLECTION_INITIAL, TAG_CHECKSUM, TAG_TYPE,
    TAG_FILENAME, TYPE_NII)
from populse_mia.data_manager.database_mia import (TAG_ORIGIN_BUILTIN,
                                                   TAG_ORIGIN_USER)

# Populse_db imports
from populse_db.database import (
    FIELD_TYPE_STRING, FIELD_TYPE_DATETIME, FIELD_TYPE_DATE,
    FIELD_TYPE_TIME, FIELD_TYPE_LIST_STRING, FIELD_TYPE_INTEGER,
    FIELD_TYPE_LIST_INTEGER, FIELD_TYPE_FLOAT, FIELD_TYPE_LIST_FLOAT,
    FIELD_TYPE_BOOLEAN, FIELD_TYPE_LIST_BOOLEAN, FIELD_TYPE_LIST_DATE,
    FIELD_TYPE_LIST_DATETIME, FIELD_TYPE_LIST_TIME)


class ImportProgress(QProgressDialog):
    """Handle the progress bar.

    Attributes;
        :param project: A Project object

    Methods:
        - onProgress : Set the import progressbar value.

    """

    def __init__(self, project):
        super(ImportProgress, self).__init__(
            "Please wait while the paths are being imported...", None, 0, 3)

        self.setWindowTitle("Importing the paths")
        self.setWindowFlags(Qt.Window | Qt.WindowTitleHint |
                            Qt.CustomizeWindowHint)
        self.setModal(True)

        self.setMinimumDuration(0)
        self.setValue(0)
        self.setMinimumWidth(350) # For mac OS

        self.worker = ImportWorker(project, self)
        self.worker.finished.connect(self.close)
        self.worker.notifyProgress.connect(self.onProgress)
        self.worker.start()

    def onProgress(self, i):
        """Signal to set the import progressbar value"""

        self.setValue(i)


class ImportWorker(QThread):
    """Manage threads.

    Attributes:
        :param project: A Project object
        ;param progress: An ImportProgress object

    Methods:
        - run : Override the QThread run method.
    """
    # Used to fill the progress bar
    notifyProgress = pyqtSignal(int)

    def __init__(self, project, progress):
        super().__init__()
        self.project = project
        self.progress = progress
        self.lock = threading.RLock()
        # scans_added should always be accessed through the lock, and copied
        # before releasing the lock, because its value will change inside
        # the thread
        self.scans_added = []

    def run(self):
        """Override the QThread run method. Executed when the worker is
        started, fills the database and updates the progress.
        """
        begin = time()

        raw_data_folder = os.path.relpath(os.path.join(self.project.folder,
                                                       'data', 'raw_data'))

        # Checking all the export logs from MRIManager and taking the most
        # recent
        list_logs = glob.glob(os.path.join(raw_data_folder, "logExport*.json"))

        if len(list_logs) == 0:
            list_dict_log = []
            
        else:
            log_to_read = max(list_logs, key=os.path.getctime)

            with open(log_to_read, "r", encoding="utf-8") as file:
                list_dict_log = json.load(file)

        # For history
        historyMaker = list()
        historyMaker.append("add_scans")
        with self.lock:
            self.scans_added = []
        values_added = []
        tags_added = []
        tags_names_added = []
        documents = {}

        # List of tags to remove
        tags_to_remove = ["Dataset data file", "Dataset header file"]

        for dict_log in list_dict_log:

            if dict_log['StatusExport'] == "Export ok":
                file_name = dict_log['NameFile']
                path_name = raw_data_folder
                
                with open(os.path.join(path_name, file_name) + ".nii",
                          'rb') as scan_file:
                    data = scan_file.read()
                    original_md5 = hashlib.md5(data).hexdigest()

                file_path = os.path.join(raw_data_folder, file_name + ".nii")
                file_database_path = os.path.relpath(file_path,
                                                     self.project.folder)

                document_not_existing = self.project.session.get_document(
                    COLLECTION_CURRENT, file_database_path) is None
                
                if document_not_existing:
                    
                    with self.lock:
                        # Scan added to history
                        self.scans_added.append(file_database_path)

                documents[file_database_path] = {}
                documents[file_database_path][TAG_FILENAME] = \
                    file_database_path

                #print('\ntags_from_file(file_name, path_name): ', tags_from_file(file_name, path_name))
                
                # For each tag in each scan 
                for tag in tags_from_file(file_name, path_name):

                    # We do the tag only if it's not in the tags to remove
                    if tag[0] not in tags_to_remove:
                        tag_name = tag[0]

                        #print('\ntag_name: ', tag_name)
                        
                        properties = tag[1]

                        #print('properties: ', properties)
                        
                        format = ''
                        description = None
                        unit = None
                        tag_type = FIELD_TYPE_STRING
                        
                        if isinstance(properties, dict):
                            format = properties['format']

                            if properties['description'] != "":
                                description = properties['description']

                            if properties['units'] != "":
                                unit = properties['units']

                            if properties['type'] != "":
                                tag_type = properties['type']
                                
                            value = properties['value']
                            
                        else:
                            value = properties[0]

                        # Creating date types
                        if format is not None and format != "":
                            format = format.replace("yyyy", "%Y")
                            format = format.replace("MM", "%m")
                            format = format.replace("dd", "%d")
                            format = format.replace("HH", "%H")
                            format = format.replace("mm", "%M")
                            format = format.replace("ss", "%S")
                            format = format.replace("SSS", "%f")
                            
                            if ("%Y" in format and "%m" in format and "%d" in
                                    format and "%H" in format and
                                    "%M" in format and "%S" in format):
                                tag_type = FIELD_TYPE_DATETIME
                                
                            elif ("%Y" in format and "%m" in format and "%d"
                                    in format):
                                tag_type = FIELD_TYPE_DATE
                                
                            elif ("%H" in format and "%M" in format and "%S"
                                    in format):
                                tag_type = FIELD_TYPE_TIME

                        if tag_name != "Json_Version":
                            # Preparing value and type

                            if ((len(value) is 1 and isinstance(value[0], list))
                              or
                               (len(value) is not 1)):

                                if tag_type == FIELD_TYPE_STRING:
                                    tag_type = FIELD_TYPE_LIST_STRING
                                    
                                elif tag_type == FIELD_TYPE_INTEGER:
                                    tag_type = FIELD_TYPE_LIST_INTEGER
                                    
                                elif tag_type == FIELD_TYPE_FLOAT:
                                    tag_type = FIELD_TYPE_LIST_FLOAT
                                    
                                elif tag_type == FIELD_TYPE_BOOLEAN:
                                    tag_type = FIELD_TYPE_LIST_BOOLEAN
                                
                                elif tag_type == FIELD_TYPE_DATE:
                                    tag_type = FIELD_TYPE_LIST_DATE
                                    
                                elif tag_type == FIELD_TYPE_DATETIME:
                                    tag_type = FIELD_TYPE_LIST_DATETIME
                                    
                                elif tag_type == FIELD_TYPE_TIME:
                                    tag_type = FIELD_TYPE_LIST_TIME 
                            
                            if len(value) is 1:
                                value = value[0]

                            else:  
                                value_prepared = []
                                
                                for value_single in value:
                                    value_prepared.append(value_single[0])
                                    
                                value = value_prepared

                        #print('value: ', value)
                        #print('tag_type: ', tag_type)

                        if (tag_type == FIELD_TYPE_DATETIME or
                            tag_type == FIELD_TYPE_DATE or
                            tag_type == FIELD_TYPE_TIME):
                            
                            if value is not None and value != "":
                                value = datetime.strptime(value, format)
                                
                                if tag_type == FIELD_TYPE_TIME:
                                    value = value.time()
                                    
                                elif tag_type == FIELD_TYPE_DATE:
                                    value = value.date()

                        # TODO time lists

                        tag_row = self.project.session.get_field(
                            COLLECTION_CURRENT, tag_name)
                        
                        if (tag_row is None and tag_name not in
                                tags_names_added):
                            # Adding the tag as it's not in the database yet
                            tags_added.append(
                                [COLLECTION_CURRENT, tag_name, tag_type,
                                 description, False, TAG_ORIGIN_BUILTIN, unit,
                                 None])
                            tags_added.append(
                                [COLLECTION_INITIAL, tag_name, tag_type,
                                 description, False, TAG_ORIGIN_BUILTIN, unit,
                                 None])
                            tags_names_added.append(tag_name)

                        # The value is accepted if it's not empty or null
                        if value is not None and value != "":
                            
                            if document_not_existing:
                                values_added.append(
                                    [file_database_path, tag_name, value,
                                     value])  # Value added to history
                            documents[file_database_path][tag_name] = value

                if document_not_existing:
                    # Tags added manually
                    # Value added to history
                    values_added.append(
                        [file_database_path, TAG_CHECKSUM, original_md5,
                         original_md5])
                    # Value added to history
                    values_added.append([file_database_path, TAG_TYPE,
                                         TYPE_NII, TYPE_NII])
                documents[file_database_path][TAG_CHECKSUM] = original_md5
                documents[file_database_path][TAG_TYPE] = TYPE_NII

        # Missing values added thanks to default values
        for tag in self.project.session.get_fields(COLLECTION_CURRENT):
            if tag.origin == TAG_ORIGIN_USER:
                for scan in self.scans_added:
                    if tag.default_value is not None and \
                            self.project.session.get_value(
                                COLLECTION_CURRENT, scan[0], tag.name) is None:
                        # Value added to history
                        values_added.append([scan, tag.name,
                                             tag.default_value,
                                             tag.default_value])
                        documents[scan][tag.name] = tag.default_value

        self.notifyProgress.emit(1)
        sleep(0.1)

        self.project.session.add_fields(tags_added)

        self.notifyProgress.emit(2)
        sleep(0.1)

        current_paths = self.project.session.get_documents_names(
            COLLECTION_CURRENT)

        for document in documents:
            if document in current_paths:
                self.project.session.remove_document(COLLECTION_CURRENT,
                                                     document)
                self.project.session.remove_document(COLLECTION_INITIAL,
                                                     document)
 
            self.project.session.add_document(COLLECTION_CURRENT, documents[
                document], flush=False)
            self.project.session.add_document(COLLECTION_INITIAL, documents[
                document], flush=False)
            
        self.project.session.session.flush()
        self.notifyProgress.emit(3)
        sleep(0.1)

        # For history
        historyMaker.append(self.scans_added)
        historyMaker.append(values_added)
        self.project.undos.append(historyMaker)
        self.project.redos.clear()

        print('\nData export duration in the database:')
        print("read_log time: " + str(round(time() - begin, 2)) + ' s\n')
        # print("read_log time: " + str(time() - begin))

        # pr.disable()
        # pr.print_stats(sort='time')
        # prof.print_stats()

def read_log(project, main_window):
    """Show the evolution of the progress bar and returns its feedback, a list
    of the paths to each data file that was loaded.

    :param project: current project in the software
    :param main_window: software's main window
    :returns: the scans that have been added
    """

    main_window.progress = ImportProgress(project)
    main_window.progress.show()
    main_window.progress.exec()

    with main_window.progress.worker.lock:
        scans_added = list(main_window.progress.worker.scans_added)
    return scans_added


# def save_project(project):
#     """
#     Saves the modifications of the project
#
#     :param project: current project in the software
#     """
#
#     project.saveModifications()

def tags_from_file(file_path, path):
    """Return a list of [tag, value] contained in a Json file.

    :param file_path: file path of the Json file (without the extension)
    :param path: project path
    :returns: a list of the Json tags of the file
    """
    json_tags = []
    with open(os.path.join(path, file_path) + ".json") as f:
        for name, value in json.load(f).items():
            json_tags.append([name, value])
    return json_tags


def verify_scans(project):
    """Check if the project's scans have been modified.

    :param project: current project in the software
    :returns: the list of scans that have been modified
    """

    # Returning the files that are problematic
    return_list = []
    for scan in project.session.get_documents_names(COLLECTION_CURRENT):

        file_name = scan
        file_path = os.path.relpath(os.path.join(project.folder, file_name))

        if os.path.exists(file_path):
            # If the file exists, we do the checksum
            with open(file_path, 'rb') as scan_file:
                data = scan_file.read()
                actual_md5 = hashlib.md5(data).hexdigest()

            initial_checksum = project.session.get_value(COLLECTION_CURRENT,
                                                         scan, TAG_CHECKSUM)
            if initial_checksum is not None and actual_md5 != initial_checksum:
                return_list.append(file_name)

        else:
            # Otherwise, we directly add the file in the list
            return_list.append(file_name)

    return return_list





# -*- coding: utf-8 -*- # Character encoding, recommended
""" Make populse_mia accessible as a module.

For example by using python3 -m populse_mia.

"""
##########################################################################
# Populse_mia - Copyright (C) IRMaGe/CEA, 2018
# Distributed under the terms of the CeCILL license, as published by
# the CEA-CNRS-INRIA. Refer to the LICENSE file or to
# http://www.cecill.info/licences/Licence_CeCILL_V2.1-en.html
# for details.
##########################################################################

# Populse_MIA imports
from populse_mia.main import main

if __name__ == '__main__':
    main()



##########################################################################
# Populse_mia - Copyright (C) IRMaGe/CEA, 2018
# Distributed under the terms of the CeCILL license, as published by
# the CEA-CNRS-INRIA. Refer to the LICENSE file or to
# http://www.cecill.info/licences/Licence_CeCILL_V2.1-en.html
# for details.
##########################################################################

import sys
import os
import yaml
import inspect
import pkgutil
from zipfile import ZipFile, is_zipfile
import tempfile
import shutil
from datetime import datetime
import distutils.dir_util
import traceback

# PyQt5 import # TO REMOVE
from PyQt5 import QtCore
from PyQt5.QtCore import QSortFilterProxyModel

# PyQt / PySide import, via soma
from soma.qt_gui import qt_backend
from soma.qt_gui.qt_backend.QtCore import (Qt, Signal, QModelIndex,
    QAbstractItemModel, QByteArray, QMimeData)
from soma.qt_gui.qt_backend.Qt import (QWidget, QTreeWidget, QLabel,
    QPushButton, QDialog, QTreeWidgetItem, QHBoxLayout, QVBoxLayout,
    QLineEdit, QApplication, QSplitter, QTreeView, QFileDialog,
    QMessageBox)

# Populse_MIA import
from populse_mia.software_properties import Config
from populse_mia.software_properties import verCmp

# capsul import
from capsul.api import get_process_instance


class ProcessLibraryWidget(QWidget):
    """
    Widget that handles the available Capsul's processes in the software

    Methods:
        - load_config: read the config in process_config.yml and return it as
          a dictionary
        - load_packages: sets packages and paths to the widget and to the
          system paths
        - open_pkg_lib: opens the package library
        - save_config: saves the current config to process_config.yml
        - update_config: updates the config and loads the corresponding
          packages
        - update_process_library: updates the tree of the process library

    """

    def __init__(self, main_window=None):
        """
        Initializes the ProcessLibraryWidget
        """

        super(ProcessLibraryWidget, self).__init__(parent=main_window)
        self.setWindowTitle("Process Library")
        self.main_window = main_window

        # Process Config
        self.update_config()

        # Process Library
        self.process_library = ProcessLibrary(self.packages)
        self.process_library.setDragDropMode(self.process_library.DragOnly)
        self.process_library.setAcceptDrops(False)
        self.process_library.setDragEnabled(True)
        self.process_library.setSelectionMode(
            self.process_library.SingleSelection)

        # # Push button to call the package library
        # push_button_pkg_lib = QPushButton()
        # push_button_pkg_lib.setText('Package library manager')
        # push_button_pkg_lib.clicked.connect(self.open_pkg_lib)

        # Test to see the inputs/outputs of a process
        self.label_test = QLabel()

        # Splitter
        self.splitter = QSplitter(Qt.Horizontal)
        self.splitter.addWidget(self.label_test)
        self.splitter.addWidget(self.process_library)

        # Layout
        h_box = QVBoxLayout()
        # h_box.addWidget(push_button_pkg_lib)
        h_box.addWidget(self.splitter)

        self.setLayout(h_box)

        self.pkg_library = PackageLibraryDialog(parent=self.main_window)
        self.pkg_library.signal_save.connect(self.update_process_library)

    def update_config(self):
        """
        Updates the config and loads the corresponding packages

        """

        self.process_config = self.load_config()
        self.load_packages()

    def update_process_library(self):
        """
        Updates the tree of the process library

        """

        self.update_config()
        self.process_library.package_tree = self.packages
        self.process_library.load_dictionary(self.packages)

    def open_pkg_lib(self):
        """
        Opens the package library

        """

        self.pkg_library.show()

    @staticmethod
    def load_config():
        """
        Read the config in process_config.yml and return it as a dictionary

        :return: the config as a dictionary
        """

        config = Config()

        if not os.path.exists(os.path.join(config.get_mia_path(), 'properties',
                                           'process_config.yml')):
            open(os.path.join(config.get_mia_path(), 'properties',
                              'process_config.yml'), 'a').close()

        with open(os.path.join(config.get_mia_path(), 'properties',
                               'process_config.yml'), 'r') as stream:

            try:
                if verCmp(yaml.__version__, '5.1', 'sup'):
                    return yaml.load(stream, Loader=yaml.FullLoader)
                else:
                    return yaml.load(stream)

            except yaml.YAMLError as exc:
                print(exc)

    def load_packages(self):
        """
        Sets packages and paths to the widget and to the system paths

        """

        try:
            self.packages = self.process_config["Packages"]
        except KeyError:
            self.packages = {}
        except TypeError:
            self.packages = {}

        try:
            self.paths = self.process_config["Paths"]
        except KeyError:
            self.paths = []
        except TypeError:
            self.paths = []

        for path in self.paths:
            # Adding the module path to the system path
            # sys.path.insert(0, os.path.abspath(path))
            if os.path.abspath(path) not in sys.path:
                sys.path.append(os.path.abspath(path))

    def save_config(self):
        """
        Saves the current config to process_config.yml

        """

        config = Config()

        self.process_config["Packages"] = self.packages
        self.process_config["Paths"] = self.paths
        with open(os.path.join(config.get_mia_path(), 'properties',
                               'process_config.yml'), 'w', encoding='utf8') \
                as stream:
            yaml.dump(self.process_config, stream, default_flow_style=False,
                      allow_unicode=True)


class Node(object):

    def __init__(self, name, parent=None):
        self._name = name
        self._parent = parent
        self._children = []
        self._value = None
        if parent is not None:
            parent.addChild(self)

    def addChild(self, child):
        self._children.append(child)

    def insertChild(self, position, child):
        if position < 0 or position > len(self._children):
            return False

        self._children.insert(position, child)
        child._parent = self
        return True

    def removeChild(self, position, child):
        if position < 0 or position > len(self._children):
            return False

        self._children.pop(position)
        child._parent = None
        return True

    def attrs(self):
        classes = self.__class__.__mro__
        keyvalued = {}
        for cls in classes:
            for key, value in cls.items():
                if isinstance(value, property):
                    keyvalued[key] = value.fget(self)
        return keyvalued

    def to_list(self):
        output = []
        if self._children:
            for child in self._children:
                output += [self.name, child.to_list()]
        else:
            output += [self.name, self.value]
        return output

    def to_dict(self, d={}):
        for child in self._children:
            child._recurse_dict(d)
        return d

    def _recurse_dict(self, d):
        if self._children:
            d[self.name] = {}
            for child in self._children:
                child._recurse_dict(d[self.name])
        else:
            d[self.name] = self.value

    def name():
        def fget(self):
            return self._name

        def fset(self, value):
            self._name = value

        return locals()

    name = property(**name())

    def value():
        def fget(self):
            return self._value

        def fset(self, value):
            self._value = value

        return locals()

    value = property(**value())

    def child(self, row):
        return self._children[row]

    def childCount(self):
        return len(self._children)

    def parent(self):
        return self._parent

    def row(self):
        if self._parent is not None:
            return self._parent._children.index(self)

    def log(self, tabLevel=-1):
        output = ''
        tabLevel += 1

        for i in range(tabLevel):
            output += '    '

        output += ''.join(('|----', self._name, ' = ', '\n'))

        for child in self._children:
            output += child.log(tabLevel)

        tabLevel -= 1
        output += '\n'
        return output

    def __repr__(self):
        return self.log()

    def data(self, column):
        if column is 0:
            parent = self._parent
            text = self.name
            while parent.name != 'Root':
                text = parent.name + '.' + text
                parent = parent._parent
            return text  # self.name

        elif column is 1:
            return self.value

    def setData(self, column, value):
        if column is 0:
            self.name = value
        if column is 1:
            self.value = value

    def resource(self):
        return None


class DictionaryTreeModel(QAbstractItemModel):
    """Data model providing a tree of an arbitrary dictionary"""

    def __init__(self, root, parent=None):
        super(DictionaryTreeModel, self).__init__(parent)
        self._rootNode = root

    def mimeTypes(self):
        return ['component/name']

    def mimeData(self, idxs):
        mimedata = QMimeData()
        for idx in idxs:
            if idx.isValid():
                node = idx.internalPointer()
                txt = node.data(idx.column())
                mimedata.setData('component/name', QByteArray(txt.encode()))
        return mimedata

    def rowCount(self, parent):
        """the number of rows is the number of children"""
        if not parent.isValid():
            parentNode = self._rootNode
        else:
            parentNode = parent.internalPointer()

        return parentNode.childCount()

    def columnCount(self, parent):
        """Number of columns is always 1 """
        return 1

    def data(self, index, role):
        """returns the data requested by the view"""
        if not index.isValid():
            return None

        node = index.internalPointer()

        if role == Qt.DisplayRole or role == Qt.EditRole:
            return node.name

    def setData(self, index, value, role=Qt.EditRole):
        """This method gets called when the user changes data"""
        if index.isValid():
            if role == Qt.EditRole:
                node = index.internalPointer()
                node.setData(index.column(), value)
                return True
        return False

    def headerData(self, section, orientation, role):
        """Return the name of the requested column"""
        if role == Qt.DisplayRole:
            if section == 0:
                return "Packages"
            if section == 1:
                return "Value"

    def flags(self, index):
        """Everything is enable and selectable. Only the leaves can be
           dragged. """
        node = index.internalPointer()
        if node.childCount() > 0:
            return Qt.ItemIsEnabled | Qt.ItemIsSelectable
        else:
            return Qt.ItemIsEnabled | Qt.ItemIsSelectable | Qt.ItemIsDragEnabled

    def parent(self, index):
        """Return the parent from given index"""
        node = self.getNode(index)
        parentNode = node.parent()
        if parentNode == self._rootNode:
            return QModelIndex()

        return self.createIndex(parentNode.row(), 0, parentNode)

    def index(self, row, column, parent):
        """Return an index from given row, column and parent"""
        parentNode = self.getNode(parent)
        childItem = parentNode.child(row)

        if childItem:
            return self.createIndex(row, column, childItem)
        else:
            return QModelIndex()

    def getNode(self, index):
        """Return a Node() from given index"""
        if index.isValid():
            node = index.internalPointer()
            if node:
                return node

        return self._rootNode

    def insertRows(self, position, rows, parent=QModelIndex()):
        """Insert rows from starting position and number given by rows"""
        parentNode = self.getNode(parent)
        self.beginInsertRows(parent, position, position + rows - 1)

        for row in range(rows):
            childCount = parentNode.childCount()
            childNode = Node("untitled" + str(childCount))
            success = parentNode.insertChild(position, childNode)

        self.endInsertRows()
        return success

    def removeRows(self, position, rows, parent=QModelIndex()):
        """Remove the rows from position to position+rows"""
        parentNode = self.getNode(parent)
        self.beginRemoveRows(parent, position, position + rows - 1)

        for row in range(rows):
            success = parentNode.removeChild(position)

        self.endRemoveRows()
        return success

    def to_dict(self):
        return self._rootNode.to_dict()


def node_structure_from_dict(datadict, parent=None, root_node=None):
    """Return a hierarchical node stucture required by the TreeModel"""
    if not parent:
        root_node = Node('Root')
        parent = root_node

    for name, data in sorted(datadict.items()):

        if isinstance(data, dict):

            if True in [True for value in data.values() if
                        value == 'process_enabled']:
                list_name = [value for value in data.values() if
                             value == 'process_enabled']

            else:

                list_name = []
                list_values = [value for value in data.values() if
                               isinstance(value, dict)]

                while (list_values):
                    value = list_values.pop()

                    for i in value.values():

                        if not isinstance(i, dict):
                            list_name.append(i)

                    list_values = list_values + [i for i in value.values() if
                                                 isinstance(i, dict)]

            # if not list_name: list_name = [i for i in data.values()]

            if (all(item == 'process_disabled' for item in list_name)):
                continue

            node = Node(name, parent)
            node = node_structure_from_dict(data, node, root_node)

        elif data == 'process_enabled':
            node = Node(name, parent)
            node.value = data

    return root_node


class ProcessLibrary(QTreeView):
    """
    Tree to display the available Capsul's processes

    Attributes:
        - dictionary: dictionary corresponding to the tree
        - _model: model used

    Methods:
        - load_dictionary: loads a dictionary to the tree
        - to_dict: returns a dictionary from the current tree

    """
    item_library_clicked = QtCore.Signal(str)

    def __init__(self, d):
        super(ProcessLibrary, self).__init__()
        self.load_dictionary(d)

    def load_dictionary(self, d):
        """
        Loads a dictionary to the tree

        :param d: dictionary to load. See the packages attribut in the
          ProcessLibraryWidget class.
        """
        self.dictionary = d
        self._nodes = node_structure_from_dict(d)
        self._model = DictionaryTreeModel(self._nodes)
        self.setModel(self._model)
        self.expandAll()

    def to_dict(self):
        """
        Returns a dictionary from the current tree

        :return: the dictionary
        """
        return self._model.to_dict()

    def mousePressEvent(self, event):
        idx = self.indexAt(event.pos())
        # print('idx',dir(idx.model()))
        if idx.isValid:
            model = idx.model()
            idx = idx.sibling(idx.row(), 0)
            node = idx.internalPointer()
            if node is not None:
                txt = node.data(idx.column())
                path = txt.encode()
                # print('dictionary ',path.decode('utf8'))
                self.item_library_clicked.emit(path.decode('utf8'))
                # self.item_library_clicked.emit(model.itemData(idx)[0])

        return QTreeView.mousePressEvent(self, event)


class PackageLibraryDialog(QDialog):
    """
    Dialog that controls which processes to show in the process library

    Attributes:
        - process_library: library of the selected processes
        - package_library: widget to control which processes to show in the
          process library
        - packages: tree-dictionary that is the representation of the
          process library
        - paths: list of path to add to the system path

    Methods:
        - load_config: updates the config and loads the corresponding packages
        - update_config: updates the process_config and package_library
          attributes
        - load_packages: updates the tree of the process library
        - save_config: saves the current config to process_config.yml
        - browse_package: opens a browser to select a package
        - add_package_with_text: adds a package from the line edit's text
        - add_package: adds a package and its modules to the package tree
        - remove_package_with_text: removes the package in the line edit from
          the package tree
        - remove_package: removes a package from the package tree
        - save: saves the tree to the process_config.yml file
        - import_file: import a python module from a path
    """

    signal_save = Signal()

    def __init__(self, parent=None):
        """
        Initialization of the PackageLibraryDialog widget
        """
        super(PackageLibraryDialog, self).__init__(parent)

        self.setWindowTitle("Package library manager")

        # True if the path specified in the line edit is a path with '/'
        self.is_path = False

        self.process_config = self.load_config()
        self.load_packages()

        self.package_library = PackageLibrary(self.packages, self.paths)

        self.status_label = QLabel()
        self.status_label.setText("")
        self.status_label.setStyleSheet(
            'QLabel{font-size:10pt;font:italic;text-align: center}')

        self.line_edit = QLineEdit()
        self.line_edit.setPlaceholderText(
            'Type a Python package (ex. nipype.interfaces.spm)')

        '''push_button_browse = QPushButton()
        push_button_browse.setText("Browse")
        push_button_browse.clicked.connect(self.browse_package)'''

        push_button_add_pkg = QPushButton()
        push_button_add_pkg.setText("Add package")
        push_button_add_pkg.clicked.connect(self.add_package_with_text)

        push_button_rm_pkg = QPushButton()
        push_button_rm_pkg.setText("Remove package")
        push_button_rm_pkg.clicked.connect(self.remove_package_with_text)

        push_button_save = QPushButton()
        push_button_save.setText("Save changes")
        push_button_save.clicked.connect(self.save)

        push_button_cancel = QPushButton("Cancel")
        push_button_cancel.setObjectName("pushButton_cancel")
        push_button_cancel.clicked.connect(self.close)

        # Layout

        h_box_line_edit = QHBoxLayout()
        h_box_line_edit.addWidget(self.line_edit)
        # h_box_line_edit.addWidget(push_button_add_pkg)
        # h_box_browse.addWidget(push_button_browse)

        h_box_label = QHBoxLayout()
        h_box_label.addStretch(1)
        h_box_label.addWidget(self.status_label)
        h_box_label.addStretch(1)

        h_box_save = QHBoxLayout()
        h_box_save.addStretch(1)
        h_box_save.addWidget(push_button_save)
        h_box_save.addWidget(push_button_cancel)

        h_box_buttons = QHBoxLayout()
        h_box_buttons.addStretch(1)
        h_box_buttons.addWidget(push_button_add_pkg)
        h_box_buttons.addStretch(1)
        h_box_buttons.addWidget(push_button_rm_pkg)
        h_box_buttons.addStretch(1)

        v_box = QVBoxLayout()
        v_box.addStretch(1)
        v_box.addLayout(h_box_label)
        v_box.addLayout(h_box_line_edit)
        v_box.addLayout(h_box_buttons)
        v_box.addStretch(1)
        v_box.addLayout(h_box_save)

        h_box = QHBoxLayout()
        h_box.addWidget(self.package_library)
        h_box.addLayout(v_box)

        self.setLayout(h_box)

    @staticmethod
    def load_config():
        """
        Updates the config and loads the corresponding packages

        :return: the config as a dictionary
        """
        config = Config()

        with open(os.path.join(config.get_mia_path(), 'properties',
                               'process_config.yml'), 'r') as stream:

            try:
                if verCmp(yaml.__version__, '5.1', 'sup'):
                    return yaml.load(stream, Loader=yaml.FullLoader)

                else:
                    return yaml.load(stream)

            except yaml.YAMLError as exc:
                print(exc)

    def update_config(self):
        """
        Updates the process_config and package_library attributes

        """
        self.process_config = self.load_config()
        self.load_packages()
        self.package_library.package_tree = self.packages
        self.package_library.paths = self.paths
        self.package_library.generate_tree()

    def load_packages(self):
        """
        Updates the tree of the process library

        """
        try:
            self.packages = self.process_config["Packages"]
        except KeyError:
            self.packages = {}
        except TypeError:
            self.packages = {}

        try:
            self.paths = self.process_config["Paths"]
        except KeyError:
            self.paths = []
        except TypeError:
            self.paths = []

    def save_config(self):
        """
        Saves the current config to process_config.yml

        """
        config = Config()
        self.process_config["Packages"] = self.packages
        self.process_config["Paths"] = self.paths
        with open(os.path.join(config.get_mia_path(), 'properties',
                               'process_config.yml'), 'w', encoding='utf8') \
                as stream:
            yaml.dump(self.process_config, stream, default_flow_style=False,
                      allow_unicode=True)

    def browse_package(self):
        """
        Opens a browser to select a package

        """
        file_dialog = QFileDialog()
        file_dialog.setOption(QFileDialog.DontUseNativeDialog, True)

        # To select files or directories, we should use a proxy model
        # but mine is not working yet...

        # file_dialog.setProxyModel(FileFilterProxyModel())
        file_dialog.setFileMode(QFileDialog.Directory)
        # file_dialog.setFileMode(QFileDialog.Directory |
        # QFileDialog.ExistingFile)
        # file_dialog.setFilter("Processes (*.py *.xml)")

        if file_dialog.exec_():
            file_name = file_dialog.selectedFiles()[0]
            file_name = os.path.abspath(file_name)
            self.is_path = True
            self.line_edit.setText(file_name)

    def add_package_with_text(self):
        """Adds a package from the line edit's text

        """
        if self.is_path:  # Currently the self.is_path = False
            # (Need to pass by the method browse_package to initialise to
            # True and the Browse button is commented.
            # Could be interesting to permit a backdoor to pass
            # the absolute path in the field for add package,
            # to be continued... )
            path, package = os.path.split(self.line_edit.text())
            # Adding the module path to the system path
            sys.path.append(path)
            self.add_package(package)
            self.paths.append(os.path.relpath(path))

        else:
            # self.package_library.package_tree = self.load_config()['Packages']
            old_status = self.status_label.text()
            self.status_label.setText(
                "Adding {0}. Please wait.".format(self.line_edit.text()))
            QApplication.processEvents()
            _2add = self.line_edit.text()

            if os.path.splitext(_2add)[1]:
                part = ''
                old_part = ''
                flag = False

                for content in _2add.split('.'):
                    part += content

                    try:
                        __import__(part)

                    except ImportError:

                        try:
                            flag = True

                            if content in dir(sys.modules[old_part]):
                                errors = self.add_package(
                                    os.path.splitext(_2add)[0],
                                    os.path.splitext(_2add)[1][1:])
                                break

                            else:
                                errors = self.add_package(_2add)
                                break

                        except KeyError:
                            errors = 'No package, module or class named ' + _2add + ' !'
                            break

                    old_part = part
                    part += '.'

                if flag is False:
                    errors = self.add_package(os.path.splitext(_2add)[0],
                                              os.path.splitext(_2add)[1][1:])

            else:
                errors = self.add_package(os.path.splitext(_2add)[0],
                                          os.path.splitext(_2add)[0])

            if len(errors) == 0:
                self.status_label.setText(
                    "{0} added to the Package Library.".format(
                        self.line_edit.text()))

            else:
                self.status_label.setText(old_status)
                msg = QMessageBox()

                if isinstance(errors, str):
                    msg.setText(errors)

                elif isinstance(errors, list):
                    msg.setText('\n'.join(errors))

                msg.setIcon(QMessageBox.Warning)
                msg.exec_()

    def add_package(self, module_name, class_name=None, show_error=False,
                    init_package_tree=False):
        """
        Adds a package and its modules to the package tree

        :param module_name: name of the module
        :param class_name: name of the class
        :param show_error: display error in a message box in case of error. If
          False, errors are silent and error messages returned at the end of
          execution
        :param init_package_tree: boolean to initialize the package tree

        """

        if init_package_tree is True:
            self.update_config()
            del self.packages

        self.packages = self.package_library.package_tree
        config = Config()

        if module_name:

            if os.path.abspath(os.path.join(config.get_mia_path(),
                                            'processes')) not in sys.path:
                sys.path.append(os.path.abspath(
                    os.path.join(config.get_mia_path(), 'processes')))

            # Reloading the package
            if module_name in sys.modules.keys():
                del sys.modules[module_name]

            err_msg = []

            try:
                __import__(module_name)
                pkg = sys.modules[module_name]

                # Checking if there are subpackages
                if hasattr(pkg, '__path__'):

                    for importer, modname, ispkg in pkgutil.iter_modules(
                            pkg.__path__):

                        if ispkg and modname != '__main__':
                            err_msg += self.add_package(
                                str(module_name + '.' + modname), class_name,
                                show_error=False)

                for k, v in sorted(list(pkg.__dict__.items())):

                    # Checking each class of in the package
                    if inspect.isclass(v):

                        try:
                            get_process_instance(
                                '%s.%s' % (module_name, v.__name__))

                        except Exception:
                            print(traceback.format_exc())
                            # TODO: WHICH TYPE OF EXCEPTION?
                            # pass

                        else:
                            # Updating the tree's dictionnary
                            path_list = module_name.split('.')
                            path_list.append(k)
                            pkg_iter = self.packages
                            recurs = False

                            for element in path_list:

                                if element == class_name:
                                    recurs = True

                                if (element in pkg_iter.keys() and element is
                                        not path_list[-1]):
                                    pkg_iter = pkg_iter[element]

                                else:

                                    if element is path_list[-1]:

                                        if (element == class_name or recurs
                                                is True):
                                            print('\nEnabling %s.%s ...' % (
                                            module_name, v.__name__))
                                            pkg_iter[
                                                element] = 'process_enabled'

                                        elif element in pkg_iter.keys():
                                            pkg_iter = pkg_iter[element]

                                        else:
                                            print(
                                                '\nA not installed pipeline '
                                                'was detected in the %s '
                                                'library:' % (path_list[0]))
                                            print('- %s.%s ...' % (
                                            module_name, v.__name__))
                                            print(
                                                'This pipeline is now '
                                                'installed but disabled '
                                                '(see File > Package Library '
                                                'Manager to enable it) ...')
                                            pkg_iter[
                                                element] = 'process_disabled'

                                    else:
                                        pkg_iter[element] = {}
                                        pkg_iter = pkg_iter[element]

                self.package_library.package_tree = self.packages
                self.package_library.generate_tree()
                # self.save_config()
                return err_msg

            except Exception as err:
                err_msg.append("in {2}: {0}: {1}.".format(err.__class__, err,
                                                          module_name))

            if show_error and len(err_msg) != 0:
                msg = QMessageBox()
                msg.setText('\n'.join(err_msg))
                msg.setIcon(QMessageBox.Warning)
                msg.exec_()

            return err_msg

        else:
            return 'No package selected!'

    def remove_package_with_text(self):
        """
        Removes the package in the line edit from the package tree

        """
        old_status = self.status_label.text()
        self.status_label.setText(
            "Removing {0}. Please wait.".format(self.line_edit.text()))
        QApplication.processEvents()
        package_removed = self.remove_package(self.line_edit.text())

        if package_removed is True:
            self.status_label.setText(
                "{0} removed from Package Library.".format(
                    self.line_edit.text()))
        else:
            self.status_label.setText(old_status)

    def remove_package(self, package):
        """
        Removes a package from the package tree

        :param package: module's representation as a string
           (e.g.: nipype.interfaces.spm)
        :return: True if the package has been removed correctly
        """
        self.packages = self.package_library.package_tree
        config = Config()

        if package:

            if os.path.abspath(os.path.join(config.get_mia_path(),
                                            'processes')) not in sys.path:
                sys.path.append(os.path.abspath(
                    os.path.join(config.get_mia_path(), 'processes')))

            path_list = package.split('.')
            pkg_iter = self.packages

            for element in path_list:

                if element in pkg_iter.keys():

                    if element is not path_list[-1]:
                        pkg_iter = pkg_iter[element]
                    else:
                        del pkg_iter[element]
                else:
                    msg = QMessageBox()
                    msg.setIcon(QMessageBox.Warning)
                    msg.setWindowTitle(
                        "Warning: Package not found in Package Library")
                    msg.setText("Package {0} not found".format(package))
                    msg.setStandardButtons(QMessageBox.Ok)
                    msg.buttonClicked.connect(msg.close)
                    msg.exec()
                    return None
                    # break

        else:
            msg = QMessageBox()
            msg.setIcon(QMessageBox.Warning)
            msg.setWindowTitle("Warning: Package not found in Package Library")
            msg.setText("No package selected!")
            msg.setStandardButtons(QMessageBox.Ok)
            msg.buttonClicked.connect(msg.close)
            msg.exec()
            return False

        self.package_library.package_tree = self.packages
        self.package_library.generate_tree()
        return True

    def save(self):
        """
        Saves the tree to the process_config.yml file

        """
        # Updating the packages and the paths according to the
        # package library tree
        self.packages = self.package_library.package_tree
        self.paths = self.package_library.paths

        if self.process_config:

            if self.process_config.get("Packages"):
                del self.process_config["Packages"]

            if self.process_config.get("Paths"):
                del self.process_config["Paths"]
        else:
            self.process_config = {}

        self.process_config["Packages"] = self.packages
        self.process_config["Paths"] = list(set(self.paths))
        config = Config()

        with open(os.path.join(config.get_mia_path(), 'properties',
                               'process_config.yml'), 'w', encoding='utf8') \
                as configfile:
            yaml.dump(self.process_config, configfile,
                      default_flow_style=False, allow_unicode=True)
            self.signal_save.emit()
        self.close()


def import_file(full_name, path):
    """
    Import a python module from a path. 3.4+ only.

    Does not call sys.modules[full_name] = path

    :param full_name: name of the package
    :param path: path of the package
    :return: the corresponding module
    """
    from importlib import util

    spec = util.spec_from_file_location(full_name, path)
    mod = util.module_from_spec(spec)

    spec.loader.exec_module(mod)
    return mod


class FileFilterProxyModel(QSortFilterProxyModel):
    """Just a test for the moment. Should be useful to use in
       the file dialog. """

    def __init__(self):
        super(FileFilterProxyModel, self).__init__()

    def filterAcceptsRow(self, source_row, source_parent):
        source_model = self.sourceModel()
        index0 = source_model.index(source_row, 0, source_parent)
        # Always show directories
        if source_model.isDir(index0):
            return True
        # filter files
        filename = source_model.fileName(index0)
        #       filename=self.sourceModel().index(row,0,parent).data().lower()
        # return True
        if filename.count(".py") + filename.count(".xml") == 0:
            return False
        else:
            return True

    # def flags(self, index):
    #     flags = super(FileFilterProxyModel, self).flags(index)
    #     source_model = self.sourceModel()
    #     if source_model.isDir(index):
    #         flags |= Qt.ItemIsSelectable
    #         return flags
    #
    #     # filter files
    #     filename = source_model.fileName(index)
    #
    #     if filename.count(".py") + filename.count(".xml") == 0:
    #         flags &= ~Qt.ItemIsSelectable
    #         return flags
    #     else:
    #         flags |= Qt.ItemIsSelectable
    #         return flags


class PackageLibrary(QTreeWidget):
    """
    Tree that displays the user-added packages and their modules
    The user can check or not each module/package

    Methods:
        - update_checks: updates the checks of the tree from an item
        - set_module_view: sets if a module has to be enabled or disabled in
          the process library
        - recursive_checks: checks/unchecks all child items
        - recursive_checks_from_child: checks/unchecks all parent items
        - generate_tree: generates the package tree
        - fill_item: fills the items of the tree recursively
    """

    def __init__(self, package_tree, paths):
        """
        Initialization of the PackageLibrary widget

        :param package_tree: representation of the packages as a tree-dictionary
        :param paths: list of paths to add to the system to import the packages
        """
        super(PackageLibrary, self).__init__()

        self.itemChanged.connect(self.update_checks)
        self.package_tree = package_tree
        self.paths = paths
        self.generate_tree()
        self.setAlternatingRowColors(True)
        self.setHeaderLabel("Packages")

    def update_checks(self, item, column):
        """
        Updates the checks of the tree from an item

        :param item: item on which to begin
        :param column: column from the check (should always be 0)
        """
        # Checked state is stored on column 0
        if column == 0:
            self.itemChanged.disconnect()
            if item.childCount():
                self.recursive_checks(item)
            if item.parent():
                self.recursive_checks_from_child(item)

            self.itemChanged.connect(self.update_checks)

    def set_module_view(self, item, state):
        """
        Sets if a module has to be enabled or disabled in the process library

        :param item: item selected in the current tree
        :param state: checked or not checked (Qt.Checked == 2.
          So if val == 2 -> checkbox is checked, and if val == 0 -> checkbox
          is not checked)
        :pkg_iter: dictionary where keys are the name of a module (brick)
            and values are 'process_enabled' or 'process_disabled'.
            Key can be a submodule. In this case the value is a dictionary
            where keys are the name of a module (brick) and values are 'process
            enabled'
            or 'process_disabled'. etc. pkg_iter take only the modules
            concerning the top
            package where a change of status where done.
        """
        if state == Qt.Checked:
            val = 'process_enabled'
        else:
            val = 'process_disabled'

        list_path = []
        list_path.append(item.text(0))
        self.top_level_items = [self.topLevelItem(i) for i in
                                range(self.topLevelItemCount())]

        while item not in self.top_level_items:
            item = item.parent()
            list_path.append(item.text(0))

        pkg_iter = self.package_tree
        list_path = list(reversed(list_path))
        for element in list_path:
            if element in pkg_iter.keys():
                if element is list_path[-1]:
                    pkg_iter[element] = val
                else:
                    pkg_iter = pkg_iter[element]
            else:
                print('Package not found')
                break

    def recursive_checks(self, parent):
        """
        Checks/unchecks all child items

        :param parent: parent item
        """
        check_state = parent.checkState(0)

        if parent.childCount() == 0:
            self.set_module_view(parent, check_state)

        for i in range(parent.childCount()):
            parent.child(i).setCheckState(0, check_state)
            self.recursive_checks(parent.child(i))

    def recursive_checks_from_child(self, child):
        """
        Checks/unchecks all parent items

        :param child: child item
        """
        check_state = child.checkState(0)

        if child.childCount() == 0:
            self.set_module_view(child, check_state)

        if child.parent():
            parent = child.parent()
            if child.checkState(0) == Qt.Checked:
                if parent.checkState(0) == Qt.Unchecked:
                    parent.setCheckState(0, Qt.Checked)
                    self.recursive_checks_from_child(parent)
            else:
                # checked_children = []
                # for child in range(parent.childCount()):
                #
                #     if child.checkState(0) == Qt.Checked:
                #         checked_children.append()

                checked_children = [child for child in
                                    range(parent.childCount())
                                    if parent.child(child).checkState(
                        0) == Qt.Checked]
                if not checked_children:
                    parent.setCheckState(0, Qt.Unchecked)
                    self.recursive_checks_from_child(parent)

    def generate_tree(self):
        """
        Generates the package tree

        """
        self.itemChanged.disconnect()
        self.clear()
        self.fill_item(self.invisibleRootItem(), self.package_tree)
        self.itemChanged.connect(self.update_checks)

    def fill_item(self, item, value):
        """
        Fills the items of the tree recursively

        :param item: current item to fill
        :param value: value of the item in the tree
        """
        item.setExpanded(True)

        if type(value) is dict:
            for key, val in sorted(value.items()):
                child = QTreeWidgetItem()
                child.setText(0, str(key))
                item.addChild(child)
                if type(val) is dict:
                    self.fill_item(child, val)
                else:
                    if val == 'process_enabled':
                        child.setCheckState(0, Qt.Checked)
                        self.recursive_checks_from_child(child)
                    elif val == 'process_disabled':
                        child.setCheckState(0, Qt.Unchecked)

        elif type(value) is list:
            for val in value:
                child = QTreeWidgetItem()
                item.addChild(child)
                if type(val) is dict:
                    child.setText(0, '[dict]')
                    self.fill_item(child, val)
                elif type(val) is list:
                    child.setText(0, '[list]')
                    self.fill_item(child, val)
                else:
                    child.setText(0, str(val))

                child.setExpanded(True)

        else:
            child = QTreeWidgetItem()
            child.setText(0, str(value))
            item.addChild(child)


class ProcessHelp(QWidget):
    ''' A widget that displays information about the selected process.
    '''

    def __init__(self, process):
        """ Generate the help.
        """
        super(ProcessHelp, self).__init__()

        label = QLabel()
        label.setText(process.help())


class InstallProcesses(QDialog):
    """A widget that allows to browse a Python package or a zip file to install
       the processes that it is containing.

    Methods:
        - get_filename: opens a file dialog to get the folder or zip file to
          install
        - install: installs the selected file/folder on Populse_MIA
    """

    process_installed = Signal()

    def __init__(self, main_window, folder):
        super(InstallProcesses, self).__init__(parent=main_window)

        self.setWindowTitle('Install processes')

        v_layout = QVBoxLayout()
        self.setLayout(v_layout)

        if folder is False:
            label_text = 'Choose zip file containing Python packages'

        elif folder is True:
            label_text = 'Choose folder containing Python packages'

        v_layout.addWidget(QLabel(label_text))

        edit_layout = QHBoxLayout()
        v_layout.addLayout(edit_layout)

        self.path_edit = QLineEdit()
        edit_layout.addWidget(self.path_edit)
        self.browser_button = QPushButton('Browse')
        edit_layout.addWidget(self.browser_button)

        bottom_layout = QHBoxLayout()
        v_layout.addLayout(bottom_layout)

        install_button = QPushButton('Install package')
        bottom_layout.addWidget(install_button)

        quit_button = QPushButton('Quit')
        bottom_layout.addWidget(quit_button)

        install_button.clicked.connect(self.install)
        quit_button.clicked.connect(self.close)
        self.browser_button.clicked.connect(
            lambda: self.get_filename(folder=folder))

    def get_filename(self, folder):
        """
        Opens a file dialog to get the folder or zip file to install

        :param folder: True if the dialog installs from folder,
          False if from zip file
        """
        if folder is True:
            filename = QFileDialog.getExistingDirectory(
                self, caption='Select a directory',
                directory=os.path.expanduser("~"),
                options=QFileDialog.ShowDirsOnly)

        elif folder is False:
            filename = QFileDialog.getOpenFileName(
                caption='Select a zip file',
                directory=os.path.expanduser("~"),
                filter='Compatible files (*.zip)')

        if filename and isinstance(filename, str):
            self.path_edit.setText(filename)

        elif filename and isinstance(filename, tuple):
            self.path_edit.setText(filename[0])

    def install(self):
        """
        Installs the selected file/folder on Populse_MIA
        """

        def add_package(proc_dic, module_name):
            """
            Adds a package and its modules to the package tree

            :param proc_dic: the process tree-dictionary
            :param module_name: name of the module
            :return: proc_dic: the modified process tree-dictionary
            """
            if module_name:

                # Reloading the package
                if module_name in sys.modules.keys():
                    del sys.modules[module_name]

                try:
                    __import__(module_name)

                except ImportError as er:
                    msg = QMessageBox()
                    msg.setIcon(QMessageBox.Critical)
                    msg.setText(('During the installation of {0}, '
                                 'the folllowing exception was raised:'
                                 '\n{1}: {2}.\nThis exception maybe '
                                 'prevented the installation ...').format(
                        module_name, er.__class__, er))
                    msg.setWindowTitle("Warning")
                    msg.setStandardButtons(QMessageBox.Ok)
                    msg.buttonClicked.connect(msg.close)
                    msg.exec()
                    raise ImportError(
                        'The {0} brick may not been installed'.format(
                            module_name))

                pkg = sys.modules[module_name]

                # Checking if there are subpackages
                for importer, modname, ispkg in pkgutil.iter_modules(
                        pkg.__path__):
                    if ispkg:
                        add_package(proc_dic, str(module_name + '.' + modname))

                for k, v in sorted(list(pkg.__dict__.items())):
                    # Checking each class of in the package
                    if inspect.isclass(v):

                        try:
                            print('\n Installing %s.%s ...' % (
                            module_name, v.__name__))
                            get_process_instance(
                                '%s.%s' % (module_name, v.__name__))

                        except Exception:
                            print(traceback.format_exc())
                            # TODO: WHICH TYPE OF EXCEPTION?
                            # pass

                        else:
                            # Updating the tree's dictionnary
                            path_list = module_name.split('.')
                            path_list.append(k)
                            pkg_iter = proc_dic

                            for element in path_list:

                                if element in pkg_iter.keys():
                                    pkg_iter = pkg_iter[element]

                                else:

                                    if element is path_list[-1]:
                                        pkg_iter[element] = 'process_enabled'

                                    else:
                                        pkg_iter[element] = {}
                                        pkg_iter = pkg_iter[element]

                return proc_dic

        def change_pattern_in_folder(path, old_pattern, new_pattern):
            """Changing all "old_pattern" pattern to "new_pattern" in the "path"
              folder
            :param path: path of the extracted or copied processes
            :param old_pattern: old pattern
            :param new_pattern: new pattern
            :return:
            """
            for dname, dirs, files in os.walk(path):

                for fname in files:
                    # Modifying only .py files (pipelines are saved with
                    # this extension)

                    if fname[-2:] == 'py':
                        fpath = os.path.join(dname, fname)

                        with open(fpath) as f:
                            s = f.read()

                        s = s.replace(old_pattern + '.', new_pattern + '.')

                        with open(fpath, "w") as f:
                            f.write(s)

        filename = self.path_edit.text()
        config = Config()

        if not os.path.isdir(filename):

            if not os.path.isfile(filename):
                msg = QMessageBox()
                msg.setIcon(QMessageBox.Critical)
                msg.setText("The specified file cannot be found")
                msg.setWindowTitle("Warning")
                msg.setStandardButtons(QMessageBox.Ok)
                msg.buttonClicked.connect(msg.close)
                msg.exec()
                return

            if os.path.splitext(filename)[1] != ".zip":
                msg = QMessageBox()
                msg.setIcon(QMessageBox.Critical)
                msg.setText("The specified file has to be a .zip file")
                msg.setWindowTitle("Warning")
                msg.setStandardButtons(QMessageBox.Ok)
                msg.buttonClicked.connect(msg.close)
                msg.exec()
                return

        try:

            if os.path.abspath(os.path.join(config.get_mia_path(),
                                            'processes')) not in sys.path:
                sys.path.append(os.path.abspath(
                    os.path.join(config.get_mia_path(),
                                 'processes')))

            # Process config update
            if not os.path.isfile(os.path.join(config.get_mia_path(),
                                               'properties',
                                               'process_config.yml')):
                open(os.path.join(config.get_mia_path(),
                                  'properties',
                                  'process_config.yml'), 'a').close()

            with open(os.path.join(config.get_mia_path(),
                                   'properties',
                                   'process_config.yml'), 'r') as stream:

                try:
                    if verCmp(yaml.__version__, '5.1', 'sup'):
                        process_dic = yaml.load(stream,
                                                Loader=yaml.FullLoader)
                    else:
                        process_dic = yaml.load(stream)

                except yaml.YAMLError as exc:
                    process_dic = {}
                    print(exc)

            if process_dic is None:
                process_dic = {}

            # Copying the original process tree
            from copy import copy
            process_dic_orig = copy(process_dic)

            try:
                packages = process_dic["Packages"]
            except KeyError:
                packages = {}
            except TypeError:
                packages = {}

            try:
                paths = process_dic["Paths"]
            except KeyError:
                paths = []
            except TypeError:
                paths = []

            # Saving all the install packages names and checking if
            # the MIA_processes are updated
            package_names = []
            mia_processes_not_found = True

            # packages_already: packages already installed in populse_
            # mia (populse_mia/processes)
            packages_already = [dire for dire in os.listdir(
                os.path.join(config.get_mia_path(), 'processes'))
                                if not os.path.isfile(
                    os.path.join(config.get_mia_path(), 'processes', dire))]

            if is_zipfile(filename):
                # Extraction of the zipped content
                with ZipFile(filename, 'r') as zip_ref:
                    packages_name = [member.split(os.sep)[0] for member in
                                     zip_ref.namelist()
                                     if (len(member.split(os.sep)) is 2 and not
                        member.split(os.sep)[-1])]

            elif os.path.isdir(
                    filename):  # !!! careful: if filename is not a zip file,
                # filename must be a directory
                # that contains only the package(s) to install!!!
                packages_name = [os.path.basename(filename)]

            for package_name in packages_name:
                # package_name: package(s) in the zip file or in folder

                if (package_name not in packages_already) or (
                        package_name == 'mia_processes'):
                    # Copy MIA_processes in a temporary folder
                    if mia_processes_not_found:

                        if (package_name == "mia_processes") and (
                                os.path.exists(
                                    os.path.join(config.get_mia_path(),
                                                 'processes',
                                                 'mia_processes'))):
                            mia_processes_not_found = False
                            tmp_folder4MIA = tempfile.mkdtemp()
                            shutil.copytree(os.path.join(config.get_mia_path(),
                                                         'processes',
                                                         'mia_processes'),
                                            os.path.join(tmp_folder4MIA,
                                                         'MIA_processes'))

                    if is_zipfile(filename):

                        with ZipFile(filename, 'r') as zip_ref:
                            members_to_extract = [member for member in
                                                  zip_ref.namelist()
                                                  if member.startswith(
                                    package_name)]
                            zip_ref.extractall(
                                os.path.join(config.get_mia_path(),
                                             'processes'), members_to_extract)

                    elif os.path.isdir(filename):
                        distutils.dir_util.copy_tree(os.path.join(filename),
                                                     os.path.join(
                                                         config.get_mia_path(),
                                                         'processes',
                                                         package_name))

                else:
                    date = datetime.now().strftime("%Y%m%d%H%M%S")

                    if is_zipfile(filename):

                        with ZipFile(filename, 'r') as zip_ref:
                            temp_dir = tempfile.mkdtemp()
                            members_to_extract = [member for member in
                                                  zip_ref.namelist()
                                                  if member.startswith(
                                    package_name)]
                            zip_ref.extractall(temp_dir, members_to_extract)
                            shutil.move(os.path.join(temp_dir, package_name),
                                        os.path.join(config.get_mia_path(),
                                                     'processes',
                                                     package_name + '_' + date))

                    elif os.path.isdir(filename):
                        shutil.copytree(os.path.join(filename),
                                        os.path.join(config.get_mia_path(),
                                                     'processes',
                                                     package_name + '_' + date))

                    original_package_name = package_name
                    package_name = package_name + '_' + date

                    # Replacing the original package name pattern in
                    # all the extracted files by the package name
                    # with the date
                    change_pattern_in_folder(
                        os.path.join(config.get_mia_path(), 'processes',
                                     package_name),
                        original_package_name, package_name)

                package_names.append(package_name)
                # package_names contains all the extracted packages
                final_package_dic = add_package(packages, package_name)

            if not os.path.abspath(
                    os.path.join(config.get_mia_path(), 'processes')) in paths:
                paths.append(os.path.abspath(
                    os.path.join(config.get_mia_path(), 'processes')))

            process_dic["Packages"] = final_package_dic
            process_dic["Paths"] = paths

            # Idea: Should we encrypt the path ?

            with open(os.path.join(config.get_mia_path(), 'properties',
                                   'process_config.yml'), 'w',
                      encoding='utf8') as stream:
                yaml.dump(process_dic, stream, default_flow_style=False,
                          allow_unicode=True)

            self.process_installed.emit()

            # Cleaning the temporary folder
            if 'tmp_folder4MIA' in locals():
                shutil.rmtree(tmp_folder4MIA)

            if 'temp_dir' in locals():
                shutil.rmtree(temp_dir)

        except Exception as e:
            # Don't know which kind of exception can be raised yet
            msg = QMessageBox()
            msg.setIcon(QMessageBox.Critical)
            msg.setText(
                '{0}: {1}\nInstallation aborted ... !'.format(e.__class__, e))
            msg.setWindowTitle("Warning")
            msg.setStandardButtons(QMessageBox.Ok)
            msg.buttonClicked.connect(msg.close)
            msg.exec()

            # Resetting process_config.yml
            if process_dic_orig is None:
                process_dic_orig = {}

            with open(os.path.join(config.get_mia_path(),
                                   'properties',
                                   'process_config.yml'),
                      'w', encoding='utf8') as stream:
                yaml.dump(process_dic_orig, stream, default_flow_style=False,
                          allow_unicode=True)

            # Deleting the extracted files
            if package_names is None:
                package_names = []

            for package_name in package_names:
                if os.path.exists(
                        os.path.join(config.get_mia_path(), 'processes',
                                     package_name)):
                    shutil.rmtree(
                        os.path.join(config.get_mia_path(), 'processes',
                                     package_name))

            # If the error comes from a MIA_process update,
            # the old version is restored
            if not mia_processes_not_found:
                distutils.dir_util.copy_tree(
                    os.path.join(tmp_folder4MIA, 'mia_processes'),
                    os.path.join(config.get_mia_path(), 'processes',
                                 'mia_processes'))

            if 'tmp_folder4MIA' in locals():
                shutil.rmtree(tmp_folder4MIA)

            if 'temp_dir' in locals():
                shutil.rmtree(temp_dir)

        else:
            msg = QMessageBox()
            msg.setWindowTitle("Installation completed")
            msg.setText("The package {0} has been correctly installed.".format(
                package_name))
            msg.setStandardButtons(QMessageBox.Ok)
            msg.buttonClicked.connect(msg.close)
            msg.exec()


if __name__ == "__main__":
    app = QApplication(sys.argv)
    print('using Qt backend:', qt_backend.get_qt_backend())
    plw = ProcessLibraryWidget()
    plw.show()
    sys.exit(app.exec_())


# -*- coding: utf-8 -*- #
"""Module that handle the configuration of the software

Load and save the parameters from the miniviewer and the MIA preferences
pop-up in the config.yml file.

Contains:
    Function:
    -verCmp
    Class:
    -Config

"""

##########################################################################
# Populse_mia - Copyright (C) IRMaGe/CEA, 2018
# Distributed under the terms of the CeCILL license, as published by
# the CEA-CNRS-INRIA. Refer to the LICENSE file or to
# http://www.cecill.info/licences/Licence_CeCILL_V2.1-en.html
# for details.
##########################################################################

import os
import yaml
import re


def verCmp(first_ver, sec_ver, comp):
    """Version comparator.

    The verCmp() function returns a boolean value to indicate whether its
    first argument (first_ver) is equal to, less or equal to, or greater or
    equal to its second argument (sec_ver), as follows:
    - if third argument (comp) is 'eq': when the first argument is equal to the
      second argument, return True (False if not).
    - if third argument (comp) is 'sup': when the first argument is greater
      than the second argument, return True (False if not).
    - if third argument (comp) is 'inf': when the first argument is less than
      the second argument, return True (False if not).

    :param first_ver: the version of a package (a string; ex. '0.13.0')
    :param sec_ver: the version of a package (a string; ex. '0.13.0')
    :param comp: comparator argument (accepted values: 'sup', 'inf' and 'eq' )

    :returns: False or True

    Contains:
        Private function:
        - normalise

     """

    def normalise(v):
        """Transform a version of a package to a corresponding list of integer.

        :returns: a list of integer (ex. [0, 13, 0])

        """

        return [int(x) for x in re.sub(r'(\.0+)*$', '', v).split(".")]

    if comp == 'eq':

        if normalise(first_ver) == normalise(sec_ver):
            return True

        else:
            return False

    elif comp == 'sup':

        if (normalise(first_ver) > normalise(sec_ver)) or (
        verCmp(first_ver, sec_ver, 'eq')):
            return True

        else:
            return False

    elif comp == 'inf':

        if (normalise(first_ver) < normalise(sec_ver)) or (
        verCmp(first_ver, sec_ver, 'eq')):
            return True

        else:
            return False


class Config:
    """
    Object that handles the configuration of the software

    Methods:
        - get_clinical_mode: returns the value of "clinical mode" checkbox
        in the preferences
        - get_matlab_command: returns Matlab command
        - get_matlab_path: returns the path of Matlab's executable
        - get_matlab_standalone_path: returns the path of Matlab Compiler
        Runtime
        - get_max_projects: returns the maximum number of projects displayed in
         the "Saved projects" menu
        - get_mia_path: returns the software's install path
        - get_projects_save_path: returns the folder where the projects are
        saved
        - get_spm_path: returns the path of SPM12 (license version)
        - get_spm_standalone_path: returns the path of SPM12 (standalone
        version)
        - get_use_matlab: returns the value of "use matlab" checkbox in the
        preferences
        - get_mri_conv_path: sets the MRIManager.jar path
        - get_use_spm: returns the value of "use spm" checkbox in the
        preferences
        - get_use_spm_standalone: returns the value of "use spm standalone"
        checkbox in the preferences
        - getBackgroundColor: returns the background color
        - getChainCursors: returns if the "chain cursors" checkbox of the
        mini viewer is activated
        - getNbAllSlicesMax: returns the maximum number of slices to display in
        the mini viewer
        - get_opened_projects: returns the opened projects
        - getPathData: returns the data's path
        - getPathToProjectsDBFile: returns the already-saved projects's path
        - getPathToProjectsFolder: returns the project's path
        - getShowAllSlices: returns if the "show all slices" checkbox of the
        mini viewer is activated
        - getTextColor: return the text color
        - getThumbnailTag: returns the tag that is displayed in the mini viewer
        - isAutoSave: checks if auto-save mode is activated
        - loadConfig: reads the config in the config.yml file
        - saveConfig: saves the config to the config.yml file
        - setAutoSave: sets the auto-save mode
        - setBackgroundColor: sets the background color
        - set_clinical_mode: sets the value of "clinical mode" checkbox in
        the preferences
        - set_matlab_path: sets the path of Matlab's executable
        - set_matlab_standalone_path: sets the path of Matlab Compiler Runtime
        - set_max_projects: sets the maximum number of projects displayed in
        the "Saved projects" menu
        - set_mia_path: sets the software's install path
        - set_mri_conv_path: sets the MRIManager.jar path
        - set_projects_save_path: sets the folder where the projects are saved
        - set_spm_path: sets the path of SPM12 (license version)
        - set_spm_standalone_path: sets the path of SPM12 (standalone version)
        - set_use_matlab: sets the value of "use matlab" checkbox in the
        preferences
        - set_use_spm: sets the value of "use spm" checkbox in the preferences
        - set_use_spm_standalone: sets the value of "use spm standalone"
        checkbox in the preferences
        - setChainCursors: sets the "chain cursors" checkbox of the mini viewer
        - setNbAllSlicesMax: sets the maximum number of slices to display in
        the mini viewer
        - set_opened_projects: sets the opened projects
        - setPathToData: sets the data's path
        - setShowAllSlices: sets the "show all slices" checkbox of the mini
        viewer
        - setTextColor: sets the text color
        - setThumbnailTag: sets the tag that is displayed in the mini viewer

    """
    def __init__(self):
        self.dev_mode = False
        self.config = self.loadConfig()
        if "mia_user_path" not in self.config.keys():
            self.config["mia_user_path"] = self.get_mia_path()
            self.saveConfig()

    def get_clinical_mode(self):
        """Get if clinical mode is disabled or enabled in the preferences.

        :return: boolean

        """
        try:
            return self.config["clinical_mode"]
        except KeyError:
            return True

    def get_matlab_command(self):
        """Get Matlab command.

        :return: Returns path to matlab executable or nothing if matlab
        path not specified

        """
        if self.config["use_spm_standalone"] == True:
            return ('{0}'.format(self.config["spm_standalone"]) + os.sep +
                    'run_spm12.sh {0}'.format(self.config[
                                                 "matlab_standalone"]) +
                    os.sep + ' script')
        elif self.config["use_matlab"] == True:
            return self.config["matlab"]
        else:
            return None

    def get_matlab_path(self):
        """Get the path to the matlab executable.

        :return: String of path

        """
        try:
            return self.config["matlab"]
        except KeyError:
            return ""

    def get_matlab_standalone_path(self):
        """Get the path to matlab compiler runtime.

        :return: string of path

        """
        try:
            return self.config["matlab_standalone"]
        except KeyError:
            return ""

    def get_max_projects(self):
        """Get the maximum number of projects displayed in the "Saved
        projects" menu.

        :return: Integer

        """
        try:
            return int(self.config["max_projects"])
        except KeyError as e:
            return 5

    def get_mia_path(self):
        """Get the path to the folder containing the processes, properties
        and resources folders of mia (mia_path).

        During the mia installation, the mia_path is defined and stored in the
        configuration.yml file, located in the .populse_mia folder (himself
        located in the user's home). If mia is launched in developer mode,
        mia path is the cloned git repository. If mia is launched in user
        mode, mia path must compulsorily be returned from the mia_path
        parameter of the configuration.yml

        :return: string of path to mia folder

        """
        dot_mia_config = os.path.join(os.path.expanduser("~"),
                                      ".populse_mia", "configuration.yml")

        if os.path.isfile(dot_mia_config):

            with open(dot_mia_config, 'r') as stream:

                try:
                    if verCmp(yaml.__version__, '5.1', 'sup'):
                        mia_home_config = yaml.load(stream,
                                                    Loader=yaml.FullLoader)
                    else:    
                        mia_home_config = yaml.load(stream)

                    if ("dev_mode" in mia_home_config.keys() and
                            mia_home_config["dev_mode"] == True):
                        # Only for developer mode
                        self.dev_mode = True
                        config_path = os.path.dirname(os.path.realpath(
                            __file__))

                        while "properties" not in os.listdir(config_path):
                            config_path = os.path.dirname(config_path)
                            
                        return config_path
                    
                    # Only for user mode
                    self.dev_mode = False

                    # mia_path is obsolete. Use mia_user_path instead
                    if "mia_path" in  mia_home_config:
                        mia_home_config["mia_user_path"] = \
                                                     mia_home_config["mia_path"]
                        del mia_home_config["mia_path"]

                        with open(dot_mia_config,
                                  'w', encoding='utf8') as configfile:
                            yaml.dump(mia_home_config, configfile,
                                      default_flow_style=False,
                                      allow_unicode=True)
                        
                    return mia_home_config["mia_user_path"]

                except (yaml.YAMLError, KeyError) as e:
                    print('\nMia path (where is located the processes, '
                          'the properties and resources folders) has not '
                          'been found ...')

        else:  # Only for developer mode
            
            try:
                return self.config["mia_user_path"]
            
            except (KeyError, AttributeError):
                config_path = os.path.dirname(os.path.realpath(
                    __file__))
                
                while "properties" not in os.listdir(config_path):
                    config_path = os.path.dirname(config_path)
                    
                return config_path

    def get_mri_conv_path(self):
        """Get the MRIManager.jar path.

        :return: string of the path

        """
        try:
            return self.config["mri_conv_path"]
        except KeyError:
            return ""
        except AttributeError:
            return ""

    def get_opened_projects(self):
        """Get opened projects.

        :return: list of opened projects

        """
        return self.config["opened_projects"]

    def get_projects_save_path(self):
        """Get the path where projects are saved.

        :return: string of path

        """
        try:
            return self.config["projects_save_path"]
        except KeyError:
            if not os.path.isdir(os.path.join(self.get_mia_path(),
                                              'projects')):
                os.mkdir(os.path.join(self.get_mia_path(), 'projects'))
            return os.path.join(self.get_mia_path(), 'projects')

    def get_spm_path(self):
        """Get the path of SPM12.

        :return: string of path

        """
        try:
            return self.config["spm"]
        except KeyError:
            return ""

    def get_spm_standalone_path(self):
        """Get the path to the SPM12 (standalone version).

        :return: String of path

        """
        try:
            return self.config["spm_standalone"]
        except KeyError:
            return ""

    def get_use_matlab(self):
        """Get the value of "use matlab" checkbox in the preferences.

        :return: boolean

        """
        try:
            return self.config["use_matlab"]
        except KeyError:
            return False

    def get_use_spm(self):
        """Get the value of "use spm" checkbox in the preferences.

        :return: boolean

        """
        try:
            return self.config["use_spm"]
        except KeyError:
            return "no"

    def get_use_spm_standalone(self):
        """Get the value of "use spm standalone" checkbox in the preferences.

        :return: boolean

        """
        try:
            return self.config["use_spm_standalone"]
        except KeyError:
            return "no"

    def getBackgroundColor(self):
        """Get background color.

        :return: string of the background color

        """
        return self.config["background_color"]

    def getChainCursors(self):
        """Get the value of the checkbox 'chain cursor' in miniviewer.

        :return: boolean

        """
        return self.config["chain_cursors"]

    def getNbAllSlicesMax(self):
        """Get number the maximum number of slices to display in the
        miniviewer.

        :return: Integer

        """
        return self.config["nb_slices_max"]

    def getPathToProjectsFolder(self):
        """Get the project's path.

        :return: string of the path

        """
        return self.config["paths"]["projects"]

    def getSourceImageDir(self):
        """Get the source directory for project images"""
        return self.config["source_image_dir"]

    def getShowAllSlices(self):
        """Get whether the show_all_slices parameters was enabled
        or not in the miniviewer.

        :return: boolean

        """
        #Used in MiniViewer
        return self.config["show_all_slices"]

    def getTextColor(self):
        """Get the text color.

        :return: string

        """
        return self.config["text_color"]

    def getThumbnailTag(self):
        """Get the tag of the thumbnail displayed in the miniviewer.

        :return: string

        """
        return self.config["thumbnail_tag"]

    def isAutoSave(self):
        """Get if the auto-save mode is enabled or not.

        :return: boolean

        """
        # used only in tests and when the background/text color is changed
        return self.config["auto_save"]

    def loadConfig(self):
        """Read the config in the config.yml file.

        :return: Returns a dictionary of the contents of config.yml

        """
        with open(os.path.join(self.get_mia_path(), 'properties',
                               'config.yml'), 'r') as stream:
            try:
                if verCmp(yaml.__version__, '5.1', 'sup'):
                    return yaml.load(stream, Loader=yaml.FullLoader)
                
                else:    
                    return yaml.load(stream)
                
            except yaml.YAMLError as exc:
                print(exc)

    def saveConfig(self):
        """Save the current parameters in the config.yml file."""

        with open(os.path.join(self.get_mia_path(), 'properties',
                               'config.yml'), 'w', encoding='utf8') as \
                configfile:

            yaml.dump(self.config, configfile, default_flow_style=False,
                      allow_unicode=True)

    def set_clinical_mode(self, clinical_mode):
        """Enable of disable clinical mode.

        :param: clinical_mode: boolean

        """
        self.config["clinical_mode"] = clinical_mode
        # Then save the modification
        self.saveConfig()

    def set_matlab_path(self, path):
        """Set the path of Matlab's executable.

        :param: path: string of path

        """
        self.config["matlab"] = path
        # Then save the modification
        self.saveConfig()

    def set_matlab_standalone_path(self, path):
        """Set the path of Matlab Compiler Runtime.

        :param: path: string of path

        """
        self.config["matlab_standalone"] = path
        # Then save the modification
        self.saveConfig()

    def set_max_projects(self, nb_max_projects):
        """Set the maximum number of projects displayed in
        the "Saved projects" menu.

        :param: nb_max_projects: Integer

        """
        self.config["max_projects"] = nb_max_projects
        # Then save the modification
        self.saveConfig()

    def set_mri_conv_path(self, path):
        """Set the MRIManager.jar path.

        :param: path: string of the path

        """
        self.config["mri_conv_path"] = path
        # Then save the modification
        self.saveConfig()

    def set_opened_projects(self, new_projects):
        """Set the list of opened projects and saves the modification.

        :param: new_projects: List of path

        """
        self.config["opened_projects"] = new_projects
        # Then save the modification
        self.saveConfig()

    def set_projects_save_path(self, path):
        """Set the folder where the projects are saved.

        :param: path: string of path
        """
        self.config["projects_save_path"] = path
        # Then save the modification
        self.saveConfig()

    def set_spm_path(self, path):
        """Set the path of SPM12 (license version).

        :param: path: string of path

        """
        self.config["spm"] = path
        # Then save the modification
        self.saveConfig()

    def set_spm_standalone_path(self, path):
        """Set the path of SPM12 (standalone version).

        :param: path: string of path

        """
        self.config["spm_standalone"] = path
        # Then save the modification
        self.saveConfig()

    def set_use_matlab(self, use_matlab):
        """Set the value of "use matlab" checkbox in the preferences.

        :param: use_matlab: boolean

        """
        self.config["use_matlab"] = use_matlab
        # Then save the modification
        self.saveConfig()

    def set_use_spm(self, use_spm):
        """Set the value of "use spm" checkbox in the preferences.

        :param: use_spm: boolean
        """
        self.config["use_spm"] = use_spm
        # Then save the modification
        self.saveConfig()

    def set_use_spm_standalone(self, use_spm_standalone):
        """Set the value of "use spm standalone" checkbox in the preferences.

        :param:use_spm_standalone: boolean

        """
        self.config["use_spm_standalone"] = use_spm_standalone
        # Then save the modification
        self.saveConfig()

    def setAutoSave(self, save):
        """Set auto-save mode.

        :param: save: boolean
        """
        self.config["auto_save"] = save
        # Then save the modification
        self.saveConfig()

    def setBackgroundColor(self, color):
        """Set background color and save configuration.

        :param: color: Color string ('Black', 'Blue', 'Green', 'Grey',
            'Orange', 'Red', 'Yellow', 'White')
        """
        self.config["background_color"] = color
        # Then save the modification
        self.saveConfig()

    def setChainCursors(self, chain_cursors):
        """Set the value of the checkbox 'chain cursor' in the miniviewer.

        :param: chain_cursors: Boolean
        """
        self.config["chain_cursors"] = chain_cursors
        # Then save the modification
        self.saveConfig()

    def setNbAllSlicesMax(self, nb_slices_max):
        """Set the number of slices to display in the miniviewer.

        :param: nb_slices_max: Int

        """
        self.config["nb_slices_max"] = nb_slices_max
        # Then save the modification
        self.saveConfig()

    def setShowAllSlices(self, show_all_slices):
        """Set the show_all_slides setting in miniviewer.

        :param: show_all_slices: Boolean
        """
        self.config["show_all_slices"] = show_all_slices
        # Then save the modification
        self.saveConfig()

    def setSourceImageDir(self, source_image_dir):
        """Set the source directory for project images"""
        self.config["source_image_dir"] = source_image_dir
        # Then save the modification
        self.saveConfig()

    def setTextColor(self, color):
        """Set text color and save configuration.

        :param: color: Color string ('Black', 'Blue', 'Green', 'Grey',
            'Orange', 'Red', 'Yellow', 'White')

        """
        self.config["text_color"] = color
        # Then save the modification
        self.saveConfig()

    def setThumbnailTag(self, thumbnail_tag):
        """Set the tag that is displayed in the mini viewer.

        :param: thumbnail_tag: String

        """
        self.config["thumbnail_tag"] = thumbnail_tag
        # Then save the modification
        self.saveConfig()

    # def set_mia_path(self, path):
    #     """
    #
    #     :param:#         path:
    #
    #     :return:
    #
    #     """
    #     self.config["mia_user_path"] = path
    #     # Then save the modification
    #     self.saveConfig()

    # def getPathData(self):
    #     """
    #     Get the path tp the data directory
    #     :return: returns the path to the data directory
    #
    #     """
    #     return self.config["paths"]["data"]

    # def getPathToProjectsDBFile(self):
    #     """
    #
    #     :return:
    #
    #     """
    #     folder = self.getPathToProjectsFolder()
    #     return os.path.join(folder, 'projects.json')

    # def setPathToData(self,path):
    #     """
    #
    #     :param:#         path:
    #
    #     :return:
    #
    #     """
    #     if path is not None and path != '':
    #         self.config["paths"]["data"] = path
    #         # Then save the modification
    #         self.saveConfig()




